@@ -352,9 +352,12 @@ class ClientImpl(Closable, Generic[T]):
                     )
                     self.unexpected_response_count += 1
                 else:
-                    fut.set_result((response, transfer))
+                    if not fut.done():  # Could have been canceled meanwhile.
+                        fut.set_result((response, transfer))
         except asyncio.CancelledError:
             _logger.debug("Cancelling the task of %s", self)
+        except pycyphal.transport.ResourceClosedError as ex:
+            _logger.debug("Cancelling the task of %s because the underlying resource is closed: %s", self, ex)
         except Exception as ex:
             exception = ex
             # Do not use f-string because it can throw, unlike the built-in formatting facility of the logger
