@@ -8,15 +8,19 @@ import errno
 import typing
 import socket
 import logging
-from ipaddress import IPv4Address, IPV4LENGTH, ip_network
+import ipaddress
+from ipaddress import IPV4LENGTH, ip_network
 import pycyphal
-from pycyphal.transport import MessageDataSpecifier, ServiceDataSpecifier, UnsupportedSessionConfigurationError
+from pycyphal.transport import MessageDataSpecifier, ServiceDataSpecifier
 from pycyphal.transport import InvalidMediaConfigurationError
 from ._socket_factory import SocketFactory, Sniffer
-from ._endpoint_mapping import SUBJECT_PORT, IP_ADDRESS_NODE_ID_MASK, service_data_specifier_to_udp_port
-from ._endpoint_mapping import node_id_to_unicast_ip, message_data_specifier_to_multicast_group
-from ._link_layer import LinkLayerCapture, LinkLayerSniffer
 
+from ._endpoint_mapping import CYPHAL_PORT
+from ._endpoint_mapping import DESTINATION_NODE_ID_MASK
+from ._endpoint_mapping import MULTICAST_PREFIX
+from ._endpoint_mapping import service_node_id_to_multicast_group, message_data_specifier_to_multicast_group
+
+from ._link_layer import LinkLayerCapture, LinkLayerSniffer
 
 _logger = logging.getLogger(__name__)
 
@@ -27,18 +31,16 @@ class IPv4SocketFactory(SocketFactory):
     a node-ID that maps to the broadcast address for the subnet is unavailable.
     """
 
-    def __init__(self, local_ip_address: IPv4Address):
-        if not isinstance(local_ip_address, IPv4Address):  # pragma: no cover
-            raise TypeError(f"Unexpected IP address type: {type(local_ip_address)}")
-        self._local = local_ip_address
+    def __init__(self, local_ip_address: ipaddress.IPv4Address):
+        self._local_ip_address = local_ip_address
 
     @property
     def max_nodes(self) -> int:
-        return IP_ADDRESS_NODE_ID_MASK  # The maximum may not be available because it may be the broadcast address.
+        return DESTINATION_NODE_ID_MASK
 
     @property
-    def local_ip_address(self) -> IPv4Address:
-        return self._local
+    def local_ip_address(self) -> ipaddress.IPv4Address:
+        return self._local_ip_address
 
     def make_output_socket(
         self, remote_node_id: typing.Optional[int], data_specifier: pycyphal.transport.DataSpecifier
@@ -52,32 +54,29 @@ class IPv4SocketFactory(SocketFactory):
             # Output sockets shall be bound, too, in order to ensure that outgoing packets have the correct
             # source IP address specified. This is particularly important for localhost; an unbound socket
             # there emits all packets from 127.0.0.1 which is certainly not what we need.
-            s.bind((str(self._local), 0))  # Bind to an ephemeral port.
+            s.bind((str(self._local_ip_address), 0))  # Bind to an ephemeral port.
         except OSError as ex:
             s.close()
             if ex.errno == errno.EADDRNOTAVAIL:
                 raise InvalidMediaConfigurationError(
-                    f"Bad IP configuration: cannot bind output socket to {self._local} [{errno.errorcode[ex.errno]}]"
+                    f"Bad IP configuration: cannot bind output socket to {self._local_ip_address} [{errno.errorcode[ex.errno]}]"
                 ) from None
             raise  # pragma: no cover
 
         if isinstance(data_specifier, MessageDataSpecifier):
-            if remote_node_id is not None:
-                s.close()
-                raise UnsupportedSessionConfigurationError("Unicast message transfers are not defined.")
+            assert remote_node_id is None  # Message transfers don't require a remote_node_id.
             # Merely binding is not enough for multicast sockets. We also have to configure IP_MULTICAST_IF.
             # https://tldp.org/HOWTO/Multicast-HOWTO-6.html
             # https://stackoverflow.com/a/26988214/1007777
-            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, self._local.packed)
+            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, self._local_ip_address.packed)
             s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, IPv4SocketFactory.MULTICAST_TTL)
-            remote_ip = message_data_specifier_to_multicast_group(self._local, data_specifier)
-            remote_port = SUBJECT_PORT
+            remote_ip = message_data_specifier_to_multicast_group(data_specifier)
+            remote_port = CYPHAL_PORT
         elif isinstance(data_specifier, ServiceDataSpecifier):
-            if remote_node_id is None:
-                s.close()
-                raise UnsupportedSessionConfigurationError("Broadcast service transfers are not defined.")
-            remote_ip = node_id_to_unicast_ip(self._local, remote_node_id)
-            remote_port = service_data_specifier_to_udp_port(data_specifier)
+            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, self._local_ip_address.packed)
+            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, IPv4SocketFactory.MULTICAST_TTL)
+            remote_ip = service_node_id_to_multicast_group(remote_node_id)
+            remote_port = CYPHAL_PORT
         else:
             assert False
 
@@ -85,7 +84,10 @@ class IPv4SocketFactory(SocketFactory):
         _logger.debug("%r: New output %r connected to remote node %r", self, s, remote_node_id)
         return s
 
-    def make_input_socket(self, data_specifier: pycyphal.transport.DataSpecifier) -> socket.socket:
+    def make_input_socket(
+        self, remote_node_id: typing.Optional[int], data_specifier: pycyphal.transport.DataSpecifier
+    ) -> socket.socket:
+        ## TODO: Add check for remote_node_id is None or not (like in make_output_socket above)
         _logger.debug("%r: Constructing new input socket for %s", self, data_specifier)
         s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         s.setblocking(False)
@@ -93,15 +95,15 @@ class IPv4SocketFactory(SocketFactory):
         # These options shall be set before the socket is bound.
         # https://stackoverflow.com/questions/14388706/how-do-so-reuseaddr-and-so-reuseport-differ/14388707#14388707
         s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        if sys.platform.startswith("linux"):  # pragma: no branch
+        if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):  # pragma: no branch
             # This is expected to be useful for unicast inputs only.
             # https://stackoverflow.com/a/14388707/1007777
             s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
 
         if isinstance(data_specifier, MessageDataSpecifier):
-            multicast_ip = message_data_specifier_to_multicast_group(self._local, data_specifier)
-            multicast_port = SUBJECT_PORT
-            if sys.platform.startswith("linux"):
+            multicast_ip = message_data_specifier_to_multicast_group(data_specifier)
+            multicast_port = CYPHAL_PORT
+            if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
                 # Binding to the multicast group address is necessary on GNU/Linux: https://habr.com/ru/post/141021/
                 s.bind((str(multicast_ip), multicast_port))
             else:
@@ -112,24 +114,33 @@ class IPv4SocketFactory(SocketFactory):
                 # Note that using INADDR_ANY in IP_ADD_MEMBERSHIP doesn't actually mean "any",
                 # it means "choose one automatically"; see https://tldp.org/HOWTO/Multicast-HOWTO-6.html
                 # This is why we have to specify the interface explicitly here.
-                s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, multicast_ip.packed + self._local.packed)
+                s.setsockopt(
+                    socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, multicast_ip.packed + self._local_ip_address.packed
+                )
             except OSError as ex:
                 s.close()
                 if ex.errno in (errno.EADDRNOTAVAIL, errno.ENODEV):
                     raise InvalidMediaConfigurationError(
-                        f"Could not register multicast group membership {multicast_ip} via {self._local} using {s} "
+                        f"Could not register multicast group membership {multicast_ip} via {self._local_ip_address} using {s} "
                         f"[{errno.errorcode[ex.errno]}]"
                     ) from None
                 raise  # pragma: no cover
         elif isinstance(data_specifier, ServiceDataSpecifier):
-            local_port = service_data_specifier_to_udp_port(data_specifier)
+            multicast_ip = service_node_id_to_multicast_group(remote_node_id)
+            multicast_port = CYPHAL_PORT
+            if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
+                s.bind((str(multicast_ip), multicast_port))
+            else:
+                s.bind(("", multicast_port))
             try:
-                s.bind((str(self._local), local_port))
+                s.setsockopt(
+                    socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, multicast_ip.packed + self._local_ip_address.packed
+                )
             except OSError as ex:
                 s.close()
                 if ex.errno in (errno.EADDRNOTAVAIL, errno.ENODEV):
                     raise InvalidMediaConfigurationError(
-                        f"Could not bind input service socket to {self._local}:{local_port} "
+                        f"Could not register multicast group membership {multicast_ip} via {self._local_ip_address} using {s} "
                         f"[{errno.errorcode[ex.errno]}]"
                     ) from None
                 raise  # pragma: no cover
@@ -139,14 +150,16 @@ class IPv4SocketFactory(SocketFactory):
         return s
 
     def make_sniffer(self, handler: typing.Callable[[LinkLayerCapture], None]) -> SnifferIPv4:
-        return SnifferIPv4(self._local, handler)
+        return SnifferIPv4(handler)
 
 
 class SnifferIPv4(Sniffer):
-    def __init__(self, local_ip_address: IPv4Address, handler: typing.Callable[[LinkLayerCapture], None]) -> None:
-        netmask_width = IPV4LENGTH - IP_ADDRESS_NODE_ID_MASK.bit_length()
-        subnet = ip_network(f"{local_ip_address}/{netmask_width}", strict=False)
-        filter_expression = f"udp and src net {subnet}"
+    def __init__(self, handler: typing.Callable[[LinkLayerCapture], None]) -> None:
+        netmask_width = IPV4LENGTH - DESTINATION_NODE_ID_MASK.bit_length() - 1  # -1 for the snm bit
+        fix = MULTICAST_PREFIX
+        subnet_ip = ipaddress.IPv4Address(fix)
+        subnet = ip_network(f"{subnet_ip}/{netmask_width}", strict=False)
+        filter_expression = f"udp and dst net {subnet}"
         _logger.debug("Constructed BPF filter expression: %r", filter_expression)
         self._link_layer = LinkLayerSniffer(filter_expression, handler)
 
@@ -155,3 +168,51 @@ class SnifferIPv4(Sniffer):
 
     def __repr__(self) -> str:
         return pycyphal.util.repr_attributes(self, self._link_layer)
+
+
+# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------
+
+
+def _unittest_udp_socket_factory_v4() -> None:
+    sock_fac = IPv4SocketFactory(local_ip_address=ipaddress.IPv4Address("127.0.0.1"))
+    assert sock_fac.local_ip_address == ipaddress.IPv4Address("127.0.0.1")
+
+    is_linux = sys.platform.startswith("linux") or sys.platform.startswith("darwin")
+
+    msg_output_socket = sock_fac.make_output_socket(remote_node_id=None, data_specifier=MessageDataSpecifier(456))
+    assert "239.0.1.200" == msg_output_socket.getpeername()[0]
+    assert CYPHAL_PORT == msg_output_socket.getpeername()[1]
+
+    srvc_output_socket = sock_fac.make_output_socket(
+        remote_node_id=123, data_specifier=ServiceDataSpecifier(456, ServiceDataSpecifier.Role.RESPONSE)
+    )
+    assert "239.1.0.123" == srvc_output_socket.getpeername()[0]
+    assert CYPHAL_PORT == srvc_output_socket.getpeername()[1]
+
+    broadcast_srvc_output_socket = sock_fac.make_output_socket(
+        remote_node_id=None, data_specifier=ServiceDataSpecifier(456, ServiceDataSpecifier.Role.RESPONSE)
+    )
+    assert "239.1.255.255" == broadcast_srvc_output_socket.getpeername()[0]
+    assert CYPHAL_PORT == broadcast_srvc_output_socket.getpeername()[1]
+
+    msg_input_socket = sock_fac.make_input_socket(remote_node_id=None, data_specifier=MessageDataSpecifier(456))
+    if is_linux:
+        assert "239.0.1.200" == msg_input_socket.getsockname()[0]
+    assert CYPHAL_PORT == msg_input_socket.getsockname()[1]
+
+    srvc_input_socket = sock_fac.make_input_socket(
+        remote_node_id=123, data_specifier=ServiceDataSpecifier(456, ServiceDataSpecifier.Role.REQUEST)
+    )
+    if is_linux:
+        assert "239.1.0.123" == srvc_input_socket.getsockname()[0]
+    assert CYPHAL_PORT == srvc_input_socket.getsockname()[1]
+
+    broadcast_srvc_input_socket = sock_fac.make_input_socket(
+        remote_node_id=None, data_specifier=ServiceDataSpecifier(456, ServiceDataSpecifier.Role.REQUEST)
+    )
+    if is_linux:
+        assert "239.1.255.255" == broadcast_srvc_input_socket.getsockname()[0]
+    assert CYPHAL_PORT == broadcast_srvc_input_socket.getsockname()[1]
+
+    sniffer = SnifferIPv4(handler=lambda x: None)
+    assert "udp and dst net 239.0.0.0/15" == sniffer._link_layer._filter_expr  # pylint: disable=protected-access
